/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>
#include <OperationsUtils.h>
#include <kernel/acl/nnfw_kernel_acl.h>
#include <arm_compute/core/Types.h>
#include <kernel/acl/Pooling.h>

#include "../util.h"

using namespace nnfw::kernel::acl;

TEST(KernelACL_TC, maxPoolFloat32_3x3to1x1)
{
  util::TensorWrapper input({1,3,3,1});
  const nnfw::rt::Shape inputShape = { OperandType::FLOAT32, {1,3,3,1}, 1.0, 0 };
  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 1;
  int32_t stride_height = 1;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  float outputData[1];
  const nnfw::rt::Shape outputShape = { OperandType::FLOAT32, {1,1,1,1}, 1.0, 0 };
  bool bret;

  float value = 1.0f;
  input.initValue([&value](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return value++;
  });

  util::initData(outputData, sizeof(outputData) / sizeof(outputData[0]), 0.0);

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bret = maxPoolFloat32(input.ptr<float>(), inputShape,
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    outputData, outputShape);
  EXPECT_EQ(bret, true);

  float expectData[] = { 9.0f };
  bret = util::compareData(outputData, expectData, outputShape);
  EXPECT_EQ(bret, true);
}

TEST(KernelACL_TC, maxPoolFloat32_3x3to1x1_RELU)
{
  util::TensorWrapper input({1,3,3,1});
  const nnfw::rt::Shape inputShape = { OperandType::FLOAT32, {1,3,3,1}, 1.0, 0 };
  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 1;
  int32_t stride_height = 1;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  float outputData[1];
  const nnfw::rt::Shape outputShape = { OperandType::FLOAT32, {1,1,1,1}, 1.0, 0 };
  bool bret;

  float value = -1.0f;
  input.initValue([&value](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return value--;
  });

  util::initData(outputData, sizeof(outputData) / sizeof(outputData[0]), 0.0);

  int32_t activation = ANEURALNETWORKS_FUSED_RELU;

  bret = maxPoolFloat32(input.ptr<float>(), inputShape,
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    outputData, outputShape);
  EXPECT_EQ(bret, true);

  float expectData[] = { 0.0f };
  bret = util::compareData(outputData, expectData, outputShape);
  EXPECT_EQ(bret, true);
}

TEST(KernelACL_TC, maxPoolFloat32_3x3to2x2)
{
  util::TensorWrapper input({1,3,3,1});
  const nnfw::rt::Shape inputShape = { OperandType::FLOAT32, {1,3,3,1}, 1.0, 0 };
  int32_t padding_left = 0;
  int32_t padding_right = 1;
  int32_t padding_top = 0;
  int32_t padding_bottom = 1;
  int32_t stride_width = 2;
  int32_t stride_height = 2;
  int32_t filter_width = 2;
  int32_t filter_height = 2;

  float outputData[4];
  const nnfw::rt::Shape outputShape = { OperandType::FLOAT32, {1,2,2,1}, 1.0, 0 };
  bool bret;

  float value = 1.0f;
  input.initValue([&value](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return value++;
  });

  util::initData(outputData, sizeof(outputData) / sizeof(outputData[0]), 0.0);

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bret = maxPoolFloat32(input.ptr<float>(), inputShape,
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    outputData, outputShape);
  EXPECT_EQ(bret, true);

  float expectData[] = {
    5.0f, 6.0f,
    8.0f, 9.0f
  };
  bret = util::compareData(outputData, expectData, outputShape);
  EXPECT_EQ(bret, true);
}

TEST(KernelACL_TC, maxPoolFloat32_147x147to73x73)
{
  util::TensorWrapper input({1,147,147,64});
  util::TensorWrapper output({1,73,73,64});

  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 2;
  int32_t stride_height = 2;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  input.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 1.0f;
  });

  output.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 0.f;
  });

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bool bret = maxPoolFloat32(input.ptr<float>(), input.shape(),
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    output.ptr<float>(), output.shape());
  EXPECT_EQ(bret, true);

  util::TensorWrapper expected({1,73,73,64});
  expected.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 1.0f;
  });

  EXPECT_EQ(output, expected);
}

TEST(KernelACL_TC, maxPoolFloat32_71x71to35x35)
{
  util::TensorWrapper input({1,71,71,192});
  util::TensorWrapper output({1,35,35,192});

  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 2;
  int32_t stride_height = 2;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  input.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 1.0f;
  });

  output.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 0.f;
  });

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bool bret = maxPoolFloat32(input.ptr<float>(), input.shape(),
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    output.ptr<float>(), output.shape());
  EXPECT_EQ(bret, true);

  util::TensorWrapper expected({1,35,35,192});
  expected.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 1.0f;
  });

  EXPECT_EQ(output, expected);
}

TEST(KernelACL_TC, averagePoolFloat32_3x3to1x1)
{
  util::TensorWrapper input({1,3,3,1});
  const nnfw::rt::Shape inputShape = { OperandType::FLOAT32, {1,3,3,1}, 1.0, 0 };
  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 1;
  int32_t stride_height = 1;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  float outputData[1];
  const nnfw::rt::Shape outputShape = { OperandType::FLOAT32, {1,1,1,1}, 1.0, 0 };
  bool bret;

  float value = 1.0f;
  input.initValue([&value](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return value++;
  });

  util::initData(outputData, sizeof(outputData) / sizeof(outputData[0]), 0.0);

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bret = averagePoolFloat32(input.ptr<float>(), inputShape,
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    outputData, outputShape);
  EXPECT_EQ(bret, true);

  float expectData[] = { 5.0f };
  bret = util::compareData(outputData, expectData, outputShape);
  EXPECT_EQ(bret, true);
}

TEST(KernelACL_TC, averagePoolFloat32_3x3to1x1_RELU)
{
  util::TensorWrapper input({1,3,3,1});
  const nnfw::rt::Shape inputShape = { OperandType::FLOAT32, {1,3,3,1}, 1.0, 0 };
  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 1;
  int32_t stride_height = 1;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  float outputData[1];
  const nnfw::rt::Shape outputShape = { OperandType::FLOAT32, {1,1,1,1}, 1.0, 0 };
  bool bret;

  float value = 3.0f;
  input.initValue([&value](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return value--;
  });

  util::initData(outputData, sizeof(outputData) / sizeof(outputData[0]), 0.0);

  int32_t activation = ANEURALNETWORKS_FUSED_RELU;

  bret = averagePoolFloat32(input.ptr<float>(), inputShape,
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    outputData, outputShape);
  EXPECT_EQ(bret, true);

  float expectData[] = { 0.0f };
  bret = util::compareData(outputData, expectData, outputShape);
  EXPECT_EQ(bret, true);
}

TEST(KernelACL_TC, averagePoolFloat32_3x3to2x2)
{
  util::TensorWrapper input({1,3,3,1});
  const nnfw::rt::Shape inputShape = { OperandType::FLOAT32, {1,3,3,1}, 1.0, 0 };
  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 1;
  int32_t stride_height = 1;
  int32_t filter_width = 2;
  int32_t filter_height = 2;

  float outputData[4];
  const nnfw::rt::Shape outputShape = { OperandType::FLOAT32, {1,2,2,1}, 1.0, 0 };
  bool bret;

  float value = 1.0f;
  input.initValue([&value](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return value++;
  });

  util::initData(outputData, sizeof(outputData) / sizeof(outputData[0]), 0.0);

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bret = averagePoolFloat32(input.ptr<float>(), inputShape,
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    outputData, outputShape);
  EXPECT_EQ(bret, true);

  float expectData[] = {
    3.0f, 4.0f,
    6.0f, 7.0f
  };  
  bret = util::compareData(outputData, expectData, outputShape);
  EXPECT_EQ(bret, true);
}

TEST(KernelACL_TC, averagePoolFloat32_3x3to3x3)
{
  std::vector<uint32_t> dims = {1,3,3,1};
  util::TensorWrapper input(dims);
  util::TensorWrapper output(dims);

  int32_t padding_left = 1;
  int32_t padding_right = 1;
  int32_t padding_top = 1;
  int32_t padding_bottom = 1;
  int32_t stride_width = 1;
  int32_t stride_height = 1;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  int32_t value=1.0f;
  input.initValue([&value](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return value++;
  });

  output.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 0.f;
  });

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bool bret = averagePoolFloat32(input.ptr<float>(), input.shape(),
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    output.ptr<float>(), output.shape());
  EXPECT_EQ(bret, true);

  util::TensorWrapper expected(dims);
  float v=2.5f;
  expected.initValue([&v](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    v = v + 0.5f;
    return v;
  });

  EXPECT_EQ(output, expected);
}

TEST(KernelACL_TC, averagePoolFloat32_35x35to35x35)
{
  int32_t N=35;
  std::vector<uint32_t> dims = {1,35,35,768};
  util::TensorWrapper input(dims);
  util::TensorWrapper output(dims);

  int32_t padding_left = 1;
  int32_t padding_right = 1;
  int32_t padding_top = 1;
  int32_t padding_bottom = 1;
  int32_t stride_width = 1;
  int32_t stride_height = 1;
  int32_t filter_width = 3;
  int32_t filter_height = 3;

  input.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 1.0f;
  });

  output.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 0.f;
  });

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bool bret = averagePoolFloat32(input.ptr<float>(), input.shape(),
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    output.ptr<float>(), output.shape());
  EXPECT_EQ(bret, true);

  util::TensorWrapper expected(dims);
  expected.initValue([&N](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
      return 1.0f;
  });

  EXPECT_EQ(output, expected);
}

TEST(KernelACL_TC, averagePoolFloat32_8x8to1x1)
{
  util::TensorWrapper input({1,8,8,2048});
  util::TensorWrapper output({1,1,1,2048});

  int32_t padding_left = 0;
  int32_t padding_right = 0;
  int32_t padding_top = 0;
  int32_t padding_bottom = 0;
  int32_t stride_width = 2;
  int32_t stride_height = 2;
  int32_t filter_width = 8;
  int32_t filter_height = 8;

  input.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 1.0f;
  });

  output.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 0.f;
  });

  int32_t activation = ANEURALNETWORKS_FUSED_NONE;

  bool bret = averagePoolFloat32(input.ptr<float>(), input.shape(),
                    padding_left, padding_right,
                    padding_top, padding_bottom,
                    stride_width, stride_height,
                    filter_width, filter_height,
                    activation,
                    output.ptr<float>(), output.shape());
  EXPECT_EQ(bret, true);

  util::TensorWrapper expected({1,1,1,2048});
  expected.initValue([](uint32_t n, uint32_t c, uint32_t h, uint32_t w) {
    return 1.0f;
  });

  EXPECT_EQ(output, expected);
}
