/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <OperationsUtils.h>
#include <NeuralNetworks.h>

#include <arm_compute/core/TensorShape.h>
#include <arm_compute/core/TensorInfo.h>
#include "../IO_accessor.h"
#include "../shape.h"
#include "../util.h"
#include "../NEUniqueTensor.h"

namespace nnfw {
namespace kernel {
namespace acl {
namespace neon {

bool softmaxFloat32(const float* inputData, const nnfw::rt::Shape& inputShape,
                    const float beta,
                    float* outputData, const nnfw::rt::Shape& outputShape)
{
  arm_compute::TensorShape input_shape = util::fromNNShape(inputShape);
  arm_compute::TensorShape output_shape = util::fromNNShape(outputShape);

  NEUniqueTensor input(arm_compute::TensorInfo(input_shape, arm_compute::Format::F32));
  NEUniqueTensor output(arm_compute::TensorInfo(output_shape, arm_compute::Format::F32));

  auto softmax_f = std::make_shared<arm_compute::NESoftmaxLayer>();
  softmax_f->configure(input.ptr(), output.ptr(), beta);

  input.allocate();
  output.allocate();

  if (inputShape.dimensions.size() == 4)
  {
    TensorAccess<InputAccessor>(input.ref(), inputData, inputShape);

    softmax_f->run();

    TensorAccess<OutputAccessor>(output.ref(), outputData, outputShape);
  }
  else if (inputShape.dimensions.size() == 2)
  {
    // Softmax comes with 1xN matrix and this is translated to N vector in arm_compute::TensorShape
    TensorAccess<VectorInputAccessor>(input.ref(), inputData, inputShape);

    softmax_f->run();

    TensorAccess<VectorOutputAccessor>(output.ref(), outputData, outputShape);
  }
  else
  {
    assert("undefined dimension of input" && 0);
    return false;
  }

  return true;
}

} // namespace neon
} // namespace acl
} // namespace kernel
} // namespace nnfw
