/*
 * Copyright (c) 2018 Samsung Electronics Co., Ltd. All Rights Reserved
 * Copyright (C) 2017 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#define LOG_TAG "Memory"

#include "Memory.h"

#include "HalInterfaces.h"
#include "Utils.h"
#include "Logging.h"

#include <unistd.h> // It's for 'close' and 'dup'
                    // TODO-NNRT : Remove this if included another header including this.
#include <cutils/ashmem.h>

namespace nnfw {
namespace rt {

int Memory::create(uint32_t size) {
    int fd = ashmem_create_region("__unused_name", size);
    if (fd == -1) {
        LOG(ERROR) << "Failed to create ashmem region.";
        return ANEURALNETWORKS_UNEXPECTED_NULL;
    }
    int prot = PROT_WRITE | PROT_READ;
    size_t offset = 0;
    void* data = mmap(0, size, prot, MAP_SHARED, fd, offset);
    if (data == MAP_FAILED) {
        LOG(ERROR) << "Can't mmap the file descriptor.";
        return ANEURALNETWORKS_UNMAPPABLE;
    } else {
        mBuffer = static_cast<uint8_t*>(data);
    }

    int dupfd = dup(fd);
    if (dupfd == -1) {
        LOG(ERROR) << "Failed to dup the fd";
        return ANEURALNETWORKS_UNEXPECTED_NULL;
    }

    if (mAshmemHandle) {
        native_handle_delete(mAshmemHandle);
    }
    mAshmemHandle = native_handle_create(1, 3);
    if (mAshmemHandle == nullptr) {
        LOG(ERROR) << "Failed to create native_handle";
        close(dupfd);
        return ANEURALNETWORKS_UNEXPECTED_NULL;
    }

    mAshmemHandle->data[0] = dupfd;
    mAshmemHandle->data[1] = prot;
    mAshmemHandle->data[2] = (int32_t)(uint32_t)(offset & 0xffffffff);
#if defined(__LP64__)
    mAshmemHandle->data[3] = (int32_t)(uint32_t)(offset >> 32);
#else
    mAshmemHandle->data[3] = 0;
#endif
    mHidlMemory = hidl_memory("mmap_fd", mAshmemHandle, size);

#if 0 // TODO-NNRT : Allocate hidl memory instead of shared memory.
    mHidlMemory = allocateSharedMemory(size);

    mMemory = mapMemory(mHidlMemory);
    if (mMemory == nullptr) {
        LOG(ERROR) << "Memory::create failed";
        return ANEURALNETWORKS_OP_FAILED;
    }
#endif // TODO-NNRT

    return ANEURALNETWORKS_NO_ERROR;
}

bool Memory::validateSize(uint32_t offset, uint32_t length) const {
    if (offset + length > mHidlMemory.size()) {
        LOG(ERROR) << "Request size larger than the memory size.";
        return false;
    } else {
        return true;
    }
}

Memory::~Memory() {
    // Delete the native_handle.
    if (mAshmemHandle) {
        int fd = mAshmemHandle->data[0];
        if (fd != -1) {
            close(fd);
        }
        native_handle_delete(mAshmemHandle);
    }
}

MemoryFd::~MemoryFd() {
    // Delete the native_handle.
    if (mHandle) {
        int fd = mHandle->data[0];
        if (fd != -1) {
            close(fd);
        }
        native_handle_delete(mHandle);
    }
}

int MemoryFd::set(size_t size, int prot, int fd, size_t offset) {
    if (fd < 0) {
        LOG(ERROR) << "ANeuralNetworksMemory_createFromFd invalid fd " << fd;
        return ANEURALNETWORKS_UNEXPECTED_NULL;
    }
    if (size == 0 || fd < 0) {
        LOG(ERROR) << "Invalid size or fd";
        return ANEURALNETWORKS_BAD_DATA;
    }
    int dupfd = dup(fd);
    if (dupfd == -1) {
        LOG(ERROR) << "Failed to dup the fd";
        return ANEURALNETWORKS_UNEXPECTED_NULL;
    }

    if (mHandle) {
        native_handle_delete(mHandle);
    }
    mHandle = native_handle_create(1, 3);
    if (mHandle == nullptr) {
        LOG(ERROR) << "Failed to create native_handle";
        close(dupfd);
        return ANEURALNETWORKS_UNEXPECTED_NULL;
    }
    mHandle->data[0] = dupfd;
    mHandle->data[1] = prot;
    mHandle->data[2] = (int32_t)(uint32_t)(offset & 0xffffffff);
#if defined(__LP64__)
    mHandle->data[3] = (int32_t)(uint32_t)(offset >> 32);
#else
    mHandle->data[3] = 0;
#endif
    mHidlMemory = hidl_memory("mmap_fd", mHandle, size);
#if 0 // TODO-NNRT : It will be opend if IMemory is added.
    return ANEURALNETWORKS_NO_ERROR;
#endif // TODO-NNRT
    return getPointer(&mBuffer);
}

int MemoryFd::getPointer(uint8_t** buffer) const {
    if (mHandle == nullptr) {
        LOG(ERROR) << "Memory not initialized";
        return ANEURALNETWORKS_UNEXPECTED_NULL;
    }

    int fd = mHandle->data[0];
    int prot = mHandle->data[1];
    size_t offset = getSizeFromInts(mHandle->data[2], mHandle->data[3]);
    void* data = mmap(nullptr, mHidlMemory.size(), prot, MAP_SHARED, fd, offset);
    if (data == MAP_FAILED) {
        LOG(ERROR) << "Can't mmap the file descriptor.";
        return ANEURALNETWORKS_UNMAPPABLE;
    } else {
        *buffer = static_cast<uint8_t*>(data);
        return ANEURALNETWORKS_NO_ERROR;
    }
}

uint32_t MemoryTracker::add(const Memory* memory) {
    VLOG(MODEL) << __func__ << " for " << memory;
    // See if we already have this memory. If so,
    // return its index.
    auto i = mKnown.find(memory);
    if (i != mKnown.end()) {
        return i->second;
    }
    VLOG(MODEL) << "It's new";
    // It's a new one.  Save it an assign an index to it.
    size_t next = mKnown.size();
    if (next > 0xFFFFFFFF) {
        LOG(ERROR) << "ANeuralNetworks more than 2^32 memories.";
        return ANEURALNETWORKS_BAD_DATA;
    }
    uint32_t idx = static_cast<uint32_t>(next);
    mKnown[memory] = idx;
    mMemories.push_back(memory);
    return idx;
}

} // namespace rt
} // namespace nnfw
